// BearLib.cpp : Defines the entry point for the console application.
//
#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <tchar.h>

#include "bearlibfov.h"

// Console-related variables
int         console_width       = 0;
int         console_height      = 0;
HANDLE      console_input       = NULL;
HANDLE      console_output      = NULL;
CHAR_INFO*  screen_buffer       = NULL;

// Current map variables
UCHAR*      map_template        = NULL; 
FOV_MAP     map;
int         sx                  = 0; // Player's startup position
int         sy                  = 0;

// Current state variables
int         px                  = 0; // Palyer's current position
int         py                  = 0;
int         visibility_radius   = 8;

bool        bricks              = false;
bool        fog_of_war          = true;

LARGE_INTEGER   pcounter_freq;

// ----------------------------------------------------------------------------

bool PollConsoleInputEvents();
INPUT_RECORD ReadConsoleKeyEvent();

bool LoadMapFromFile( LPCTSTR filename )
{
    HANDLE file = CreateFile
    (
        filename,
        GENERIC_READ,
        FILE_SHARE_READ,
        NULL,
        OPEN_EXISTING,
        FILE_ATTRIBUTE_NORMAL,
        NULL
    );

    if ( file == INVALID_HANDLE_VALUE )
    {
        return false;
    }

    DWORD size = GetFileSize( file, NULL );
    unsigned char *buffer = new unsigned char[size];

    DWORD offset = 0;
    DWORD read = 0;
    while ( offset < size )
    {
        ReadFile( file, (LPVOID)(buffer+offset), size, &read, NULL );
        offset += read;
    }

    CloseHandle( file );

    // Calculate map size
    int max_line_width = 0;
    int line_width = 0;
    int line_count = 0;

    for ( unsigned int i=0; i<size; i++ )
    {
        if ( buffer[i] == '\r' || buffer[i] == '\n' || i == size-1 ) // CR+LF (win) or LF (unix) or CR (mac)
        {
            if ( max_line_width < line_width ) max_line_width = line_width;
            line_width = 0;

            if ( (i+1 < size) && (buffer[i+1] == '\n') ) i += 1; // CR+LF (win)

            line_count += 1;
        }
        else
        {
            line_width += 1;
        }
    }

    // And through the buffer once again
    map.Width = max_line_width;
    map.Height = line_count;
    map_template = new unsigned char[map.Width * map.Height];
    for ( int i=0; i<(map.Width * map.Height); i++ ) map_template[i] = '.';

    int x = 0, y = 0;
    for ( unsigned int i=0; i<size; i++ )
    {
        if ( buffer[i] == '\r' || buffer[i] == '\n' ) // CR+LF (win) or LF (unix) or CR (mac)
        {
            if ( (i+1 < size) && (buffer[i+1] == '\n') ) i += 1; // CR+LF (win)
            x = 0;
            y += 1;
        }
        else
        {
            map_template[y*map.Width + x++] = buffer[i];
        }
    }

    map.Cells = new FOV_CELL[map.Width * map.Height];
    ZeroMemory( map.Cells, sizeof(FOV_CELL) * map.Width * map.Height );

    for ( int i=0; i < map.Width * map.Height; i++ )
    {
        switch ( map_template[i] )
        {
            case '.':
                map.Cells[i].Flags &= (0xFF ^ FOV_CELL_OPAQUE);
                break;

            case '#':
                map.Cells[i].Flags |= FOV_CELL_OPAQUE;
                break;

            case '@':
                map.Cells[i].Flags &= (0xFF ^ FOV_CELL_OPAQUE);
                sx = i%map.Width;
                sy = (i-sx)/map.Width;
                break;
        };
    }

    return true;
}

void DrawScreen()
{
    COORD buffer_size = { console_width, console_height }, position = { 0, 0 };
    SMALL_RECT rect = { 0, 0, buffer_size.X, buffer_size.Y };

    LARGE_INTEGER start, end;
    QueryPerformanceCounter( &start );

    CalcFOVC( &map, px, py, visibility_radius );

    QueryPerformanceCounter( &end );
    double time = (double)(end.QuadPart - start.QuadPart) / (double)pcounter_freq.QuadPart;

    for ( int y=0; y<map.Height; y++ )
	{
        for ( int x=0; x<map.Width; x++ )
        {
            int map_index = y*map.Width + x;
            unsigned char temp = map_template[map_index];

            unsigned char symbol = ' ';
            WORD foreground = 0;
            WORD background = 0;
        
            if ( temp == '.' || temp == '@' )
            {
                symbol = (bricks ? 0xB0 : '.');
                foreground = FOREGROUND_GREEN|FOREGROUND_RED|FOREGROUND_BLUE;
            }
            else if ( temp == '#' )
            {
                symbol = (bricks ? 0xDB : '#');
                foreground = FOREGROUND_GREEN|FOREGROUND_RED|FOREGROUND_BLUE;
            }

            if ( map.Cells[map_index].Flags & FOV_CELL_VISIBLE )
            {
                foreground = FOREGROUND_GREEN;
                foreground |= FOREGROUND_INTENSITY;
                if ( bricks ) background |= BACKGROUND_INTENSITY;
            }

            // Yet if it is still hidden by the fog of war
            if ( fog_of_war && !(map.Cells[map_index].Flags & FOV_CELL_VISITED) )
            {
                foreground = 0;
                background = 0;
            }

            int screen_index = y*console_width + x;
            screen_buffer[screen_index].Char.AsciiChar = symbol;
            screen_buffer[screen_index].Attributes = foreground | background;
        }
	}

    screen_buffer[map.Width*py + px].Char.AsciiChar = (bricks ? 0xDB : '@');
    screen_buffer[map.Width*py+px].Attributes = FOREGROUND_GREEN|FOREGROUND_RED|FOREGROUND_INTENSITY;

    // Bottom text panel
    int panel_from = map.Height*map.Width;
    int panel_to = console_height*map.Width;
    for ( int i=panel_from; i<panel_to; i++ )
    {
        screen_buffer[i].Char.AsciiChar = ' ';
        screen_buffer[i].Attributes = 0;
    }

	WriteConsoleOutputA
	(
		console_output,
		screen_buffer,
		buffer_size,
		position,
		&rect
	);
    
    COORD pos = { 1, console_height-2 };
    SetConsoleCursorPosition( console_output, pos );

    printf( "FOV took %.4f s", time );
}

unsigned char GetMapTemplateTile( int x, int y )
{
    return map_template[y*map.Width + x];
}

int main(int argc, char* argv[])
{
    QueryPerformanceFrequency( &pcounter_freq );

    console_output = GetStdHandle(STD_OUTPUT_HANDLE);
	console_input = GetStdHandle(STD_INPUT_HANDLE);

    // Make cursor invisible
	CONSOLE_CURSOR_INFO cursorInfo = { /* dwSize */ 1, /* bVisible */ FALSE };
	SetConsoleCursorInfo( console_output, &cursorInfo );

    // We need a map to set up window size
    TCHAR path[MAX_PATH];
    GetModuleFileName( 0, path, MAX_PATH );
    for ( int i = _tcslen(path); i >= 0; i-- )
    {
        if ( path[i] == (TCHAR)'\\' ) { path[i] = 0; break; }
    }
    //_tcscat_s<MAX_PATH>( path, TEXT("\\map.txt") );
    _tcscat( path, TEXT("\\map.txt") );

    if ( !LoadMapFromFile( path ) )
    {
        return 1;
    }

    // Arbitrary buffer size (disables usual scrolling)
    console_width = map.Width;
    if ( console_width < 80 ) console_width = 80;
    
    console_height = map.Height + 3;
    if ( console_height < 20 ) console_height = 20;
	
    COORD buffer_size = { console_width, console_height };
    screen_buffer = new CHAR_INFO[buffer_size.X * buffer_size.Y];
	SetConsoleScreenBufferSize( console_output, buffer_size );

    SMALL_RECT rect = { 0, 0, buffer_size.X-1, buffer_size.Y-1 };
    SetConsoleWindowInfo( console_output, TRUE, &rect );

    // Nice window title instead of .exe's path
	SetConsoleTitle( TEXT("BeaRLibFOV 11i16 \\o/") );

	px = sx;
    py = sy;

    short key_directions[][3] = 
    {
        { VK_LEFT,      -1,  0 },
        { VK_RIGHT,     +1,  0 },
        { VK_UP,         0, -1 },
        { VK_DOWN,       0, +1 },
        { VK_NUMPAD1,   -1, +1 },
        { VK_NUMPAD2,    0, +1 },
        { VK_NUMPAD3,   +1, +1 },
        { VK_NUMPAD4,   -1,  0 },
        { VK_NUMPAD6,   +1,  0 },
        { VK_NUMPAD7,   -1, -1 },
        { VK_NUMPAD8,    0, -1 },
        { VK_NUMPAD9,   +1, -1 }
    };

    DrawScreen();

	while ( true )
	{
		if ( !PollConsoleInputEvents() )
		{
			Sleep( 5 );
		}
		else
		{
			INPUT_RECORD record = ReadConsoleKeyEvent();
            if ( record.EventType == 0 ) continue;

			WORD key_code = record.Event.KeyEvent.wVirtualKeyCode;
            if ( key_code == VK_ESCAPE ) break;

            switch ( key_code )
            {
                case 'B':
                    bricks = !bricks;
                    break;

                case 'F':
                    fog_of_war = !fog_of_war;
                    break;

                case VK_ADD:
                    if ( visibility_radius < map.Width ) visibility_radius += 1;
                    break;

                case VK_SUBTRACT:
                    if ( visibility_radius > 0 ) visibility_radius -= 1;
                    break;
            };

            COORD direction = { 0, 0 };
            for ( int i=0; i<(sizeof(key_directions)/sizeof(key_directions[0])); i++ )
            {
                if (key_code == key_directions[i][0])
                {
                    direction.X = key_directions[i][1];
                    direction.Y = key_directions[i][2];
                    break;
                }
            }

            int new_px = px + direction.X;
            int new_py = py + direction.Y;

            if ( (new_px < 0) || (new_px >= map.Width) ) new_px = px;
            if ( (new_py < 0) || (new_py >= map.Height) ) new_py = py;

            if ( GetMapTemplateTile( new_px, new_py ) != '#' )
            {
                px = new_px;
                py = new_py;
            }

			DrawScreen();
		}
	}

    if ( map_template != NULL ) delete[] map_template;
    if ( screen_buffer != NULL ) delete[] screen_buffer;
    if ( map.Cells != NULL ) delete[] map.Cells;

	CloseHandle( console_output );
    CloseHandle( console_input );

	return 0;
}

// ----------------------------------------------------------------------------

bool PollConsoleInputEvents()
{
	DWORD number_of_events = 0;
	GetNumberOfConsoleInputEvents( console_input, &number_of_events );
	return (number_of_events > 0);
}

INPUT_RECORD ReadConsoleKeyEvent()
{
	INPUT_RECORD record;
	DWORD number_of_events = 0;
	do
	{
		ReadConsoleInput( console_input, &record, 1, &number_of_events );
		if ( record.EventType == KEY_EVENT &&
			 record.Event.KeyEvent.bKeyDown == TRUE )
		{
			break;
		}
	}
	while ( number_of_events > 0 );

    if ( number_of_events == 0 ) ZeroMemory( &record, sizeof(INPUT_RECORD) );
	
    return record;
}

